@@ -6,8 +6,7 @@ Class for representing record structure transformations.
 import re
 from typing import Iterable, List, Optional
 
-import numpy
-import data_algebra.pandas_model
+import data_algebra.data_model
 import data_algebra.util
 
 
@@ -16,14 +15,18 @@ class RecordSpecification:
     Class to represent a multi-row data record.
     """
 
+    row_columns: List[str]  # columns when in row form
+    block_columns: List[str]  # columns when in block form
+    strict: bool  # do or do not allow repeated value names
+
     def __init__(
         self,
         control_table,
         *,
         record_keys=None,
         control_table_keys=None,
-        strict=False,
-        local_data_model=None
+        strict: bool = True,
+        local_data_model=None,
     ):
         """
         :param control_table: data.frame describing record layout
@@ -31,23 +34,27 @@ class RecordSpecification:
                defaults to no columns.
         :param control_table_keys: array of control_table key column names,
                defaults to first column for non-trivial blocks and no columns for rows.
-        :param strict: logical, if True more checks on transform
+        :param strict: if True don't allow duplicate value names
         :param local_data_model: data.frame data model
         """
+        assert isinstance(strict, bool)
+        self.strict = strict
         if local_data_model is None:
-            local_data_model = data_algebra.pandas_model.default_data_model
-        control_table = control_table.reset_index(inplace=False, drop=True)
+            local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(control_table)
+        assert isinstance(local_data_model, data_algebra.data_model.DataModel)
+        assert local_data_model.is_appropriate_data_instance(control_table)
+        control_table = local_data_model.clean_copy(control_table)
         if control_table.shape[0] < 1:
             raise ValueError("control table should have at least 1 row")
         if len(control_table.columns) != len(set(control_table.columns)):
             raise ValueError("control table columns should be unique")
-        self.control_table = control_table.reset_index(drop=True, inplace=False)
+        self.control_table = control_table
         assert self.control_table.shape[0] > 0
         if record_keys is None:
             record_keys = []
         if isinstance(record_keys, str):
             record_keys = [record_keys]
-        self.record_keys = [k for k in record_keys]
+        self.record_keys = list(record_keys)
         if control_table_keys is None:
             if self.control_table.shape[0] > 1:
                 control_table_keys = [self.control_table.columns[0]]
@@ -55,7 +62,7 @@ class RecordSpecification:
                 control_table_keys = []  # single row records don't need to be keyed
         if isinstance(control_table_keys, str):
             control_table_keys = [control_table_keys]
-        if strict and (self.control_table.shape[0] > 1):
+        if self.control_table.shape[0] > 1:
             if len(control_table_keys) <= 0:
                 raise ValueError(
                     "multi-row records must have at least one control table key"
@@ -76,34 +83,43 @@ class RecordSpecification:
         for ck in self.control_table_keys:
             if any(local_data_model.bad_column_positions(control_table[ck])):
                 raise ValueError("NA/NaN/inf/None not allowed as control table keys")
-        if strict:
-            if not data_algebra.util.table_is_keyed_by_columns(
-                self.control_table, self.control_table_keys
-            ):
-                raise ValueError("control table wasn't keyed by control table keys")
+        if not local_data_model.table_is_keyed_by_columns(
+            self.control_table, column_names=self.control_table_keys
+        ):
+            raise ValueError("control table wasn't keyed by control table keys")
         self.block_columns = self.record_keys + [c for c in self.control_table.columns]
         cvs = []
-        for c in self.control_table:
+        for c in self.control_table.columns:
             if c not in self.control_table_keys:
                 col = self.control_table[c]
-                isnull = col.isnull()
-                if all(isnull):
-                    raise ValueError("column " + c + " was all null")
+                isnull = local_data_model.bad_column_positions(col)
+                if any(isnull):
+                    raise ValueError("column " + c + " has null(s)")
                 for i in range(len(col)):
-                    if not isnull[i]:
-                        v = col[i]
-                        if v not in cvs:
-                            cvs.append(v)
+                    v = col[i]
+                    assert isinstance(v, str)
+                    assert len(v) > 0
+                    cvs.append(v)
         confused = set(record_keys).intersection(cvs)
         if len(confused) > 0:
             raise ValueError(
                 "control table entries confused with row keys or control table keys"
             )
-        if strict:
-            if len(set(cvs)) != len(cvs):
+        if len(set(cvs)) != len(cvs):
+            if strict:
                 raise ValueError("duplicate content keys")
+            cvs_orig = cvs
+            cvs = []
+            cvs_seen = set()
+            for v in cvs_orig:
+                if v not in cvs_seen:
+                    cvs.append(v)
+                    cvs_seen.add(v)
         self.content_keys = cvs
+        # columns in the row-record form
         self.row_columns = self.record_keys + cvs
+        # columns in the block-record form
+        self.block_columns = self.record_keys + list(self.control_table.columns)
 
     def row_version(self, *, include_record_keys: bool = True) -> List[str]:
         """
@@ -127,6 +143,8 @@ class RecordSpecification:
             + data_algebra.util.pandas_to_example_str(self.control_table)
             + ",\n    control_table_keys="
             + self.control_table_keys.__repr__()
+            + ",\n    strict="
+            + self.strict.__repr__()
             + ")"
         )
         return s
@@ -167,7 +185,7 @@ class RecordSpecification:
         :return: RecordMap
         """
 
-        return RecordMap(blocks_in=self)
+        return RecordMap(blocks_in=self, strict=self.strict)
 
     def map_from_rows(self):
         """
@@ -176,181 +194,7 @@ class RecordSpecification:
         :return: RecordMap
         """
 
-        return RecordMap(blocks_out=self)
-
-
-def blocks_to_rowrecs(data, *, blocks_in: RecordSpecification, local_data_model=None):
-    """
-    Convert a block record (record spanning multiple rows) into a rowrecord (record in a single row).
-
-    :param data: data frame to be transformed
-    :param blocks_in: record specification
-    :param local_data_model: pandas model.
-    :return: transformed data frame
-    """
-    assert isinstance(blocks_in, RecordSpecification)
-    ck = [k for k in blocks_in.content_keys if k is not None]
-    if len(ck) != len(set(ck)):
-        raise ValueError("blocks_in can not have duplicate content keys")
-    if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
-    data = data.reset_index(drop=True)
-    missing_cols = set(blocks_in.control_table_keys).union(blocks_in.record_keys) - set(
-        data.columns
-    )
-    if len(missing_cols) > 0:
-        raise KeyError("missing required columns: " + str(missing_cols))
-    # table must be keyed by record_keys + control_table_keys
-    if not data_algebra.util.table_is_keyed_by_columns(
-        data, blocks_in.record_keys + blocks_in.control_table_keys
-    ):
-        raise ValueError(
-            "table is not keyed by blocks_in.record_keys + blocks_in.control_table_keys"
-        )
-    # convert to row-records
-    # regularize/complete records
-    dtemp = data.copy()  # TODO: select down columns
-    dtemp["FALSE_AGG_KEY"] = 1
-    if len(blocks_in.record_keys) > 0:
-        ideal = dtemp[blocks_in.record_keys + ["FALSE_AGG_KEY"]].copy()
-        res = ideal.groupby(blocks_in.record_keys)["FALSE_AGG_KEY"].agg("sum")
-        ideal = local_data_model.data_frame(res).reset_index(drop=False)
-        ideal["FALSE_AGG_KEY"] = 1
-        ctemp = blocks_in.control_table[blocks_in.control_table_keys].copy()
-        ctemp["FALSE_AGG_KEY"] = 1
-        ideal = ideal.merge(ctemp, how="outer", on="FALSE_AGG_KEY")
-        ideal = ideal.reset_index(drop=True)
-        dtemp = ideal.merge(
-            right=dtemp,
-            how="left",
-            on=blocks_in.record_keys + blocks_in.control_table_keys + ["FALSE_AGG_KEY"],
-        )
-    dtemp.sort_values(
-        by=blocks_in.record_keys + blocks_in.control_table_keys, inplace=True
-    )
-    dtemp = dtemp.reset_index(drop=True)
-    # start building up result frame
-    if len(blocks_in.record_keys) > 0:
-        res = dtemp.groupby(blocks_in.record_keys)["FALSE_AGG_KEY"].agg("sum")
-    else:
-        res = dtemp.groupby("FALSE_AGG_KEY")["FALSE_AGG_KEY"].agg("sum")
-    res = local_data_model.data_frame(res).reset_index(drop=False)
-    res.sort_values(by=blocks_in.record_keys, inplace=True)
-    res = local_data_model.data_frame(res).reset_index(drop=True)
-    del res["FALSE_AGG_KEY"]
-    # now fill in columns
-    ckeys = blocks_in.control_table_keys
-    value_keys = [k for k in blocks_in.control_table.columns if k not in set(ckeys)]
-    donor_cols = set(dtemp.columns)
-    for i in range(blocks_in.control_table.shape[0]):
-        want = numpy.ones((dtemp.shape[0],), dtype=bool)
-        for ck in ckeys:
-            want = numpy.logical_and(want, dtemp[ck] == blocks_in.control_table[ck][i])
-        if numpy.any(want):
-            for vk in value_keys:
-                if vk in donor_cols:
-                    dcol = blocks_in.control_table[vk][i]
-                    res[dcol] = numpy.asarray(dtemp.loc[want, vk])
-    # fill in any missed columns
-    colset = set(res.columns)
-    for c in blocks_in.row_version():
-        if c not in colset:
-            res[c] = None
-    if data.shape[0] <= 0:
-        res = res.loc[range(0), :]
-        res = res.reset_index(inplace=False, drop=True)
-    return res
-
-
-def rowrecs_to_blocks(
-    data,
-    *,
-    blocks_out: RecordSpecification,
-    check_blocks_out_keying: bool = False,
-    local_data_model=None
-):
-    """
-    Convert rowrecs (single row records) into block records (multiple row records).
-
-    :param data: data frame to transform.
-    :param blocks_out: record specification.
-    :param check_blocks_out_keying: logical, if True confirm keying
-    :param local_data_model: pandas data model
-    :return: transformed data frame
-    """
-    assert isinstance(blocks_out, RecordSpecification)
-    if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
-    data = data.reset_index(drop=True)
-    missing_cols = set(blocks_out.record_keys) - set(data.columns)
-    if len(missing_cols) > 0:
-        raise KeyError("missing required columns: " + str(missing_cols))
-    if check_blocks_out_keying:
-        # prefer table be keyed by record_keys
-        if not data_algebra.util.table_is_keyed_by_columns(
-            data, blocks_out.record_keys
-        ):
-            raise ValueError("table is not keyed by blocks_out.record_keys")
-    # convert to block records, first build up parallel structures
-    rv = [k for k in blocks_out.row_version(include_record_keys=True) if k is not None]
-    if len(rv) != len(set(rv)):
-        raise ValueError("duplicate row columns")
-    dtemp_cols = [
-        k
-        for k in rv
-        if k is not None and k in set(blocks_out.record_keys + blocks_out.content_keys)
-    ]
-    dtemp = data[dtemp_cols].copy()
-    dtemp.sort_values(by=blocks_out.record_keys, inplace=True)
-    dtemp = dtemp.reset_index(drop=True)
-    if len(dtemp.columns) != len(set(dtemp.columns)):
-        raise ValueError("targeted data columns not unique")
-    ctemp = blocks_out.control_table.copy()
-    dtemp["FALSE_JOIN_KEY"] = 1
-    ctemp["FALSE_JOIN_KEY"] = 1
-    res = dtemp[blocks_out.record_keys + ["FALSE_JOIN_KEY"]].merge(
-        ctemp, how="outer", on=["FALSE_JOIN_KEY"]
-    )
-    del res["FALSE_JOIN_KEY"]
-    ckeys = blocks_out.control_table_keys
-    res.sort_values(by=blocks_out.record_keys + ckeys, inplace=True)
-    res = res.reset_index(drop=True)
-    del ctemp["FALSE_JOIN_KEY"]
-    del dtemp["FALSE_JOIN_KEY"]
-    value_keys = [k for k in ctemp.columns if k not in set(ckeys)]
-    donor_cols = set(dtemp.columns)
-    for vk in value_keys:
-        res[vk] = None
-    # we now have parallel structures to copy between
-    for i in range(ctemp.shape[0]):
-        want = numpy.ones((res.shape[0],), dtype=bool)
-        for ck in ckeys:
-            want = numpy.logical_and(want, res[ck] == ctemp[ck][i])
-        if numpy.any(want):
-            for vk in value_keys:
-                dcol = ctemp[vk][i]
-                if dcol in donor_cols:
-                    nvals = numpy.asarray(dtemp[dcol])
-                    if len(nvals) < 1:
-                        nvals = [None] * numpy.sum(want)
-                    if numpy.all(want):
-                        res[vk] = nvals  # get around Pandas future warning
-                    else:
-                        res.loc[want, vk] = nvals
-    # see about promoting composite columns to numeric
-    for vk in set(value_keys):
-        converted = local_data_model.to_numeric(res[vk], errors="coerce")
-        if numpy.all(
-            local_data_model.isnull(converted) == local_data_model.isnull(res[vk])
-        ):
-            res[vk] = converted
-    if data.shape[0] < 1:
-        # empty input produces emtpy output (with different column structure)
-        res = res.iloc[range(0), :].reset_index(inplace=False, drop=True)
-    if data.shape[0] <= 0:
-        res = res.loc[range(0), :]
-        res = res.reset_index(inplace=False, drop=True)
-    return res
+        return RecordMap(blocks_out=self, strict=self.strict)
 
 
 class RecordMap:
@@ -358,25 +202,37 @@ class RecordMap:
     Class for specifying general record to record transforms.
     """
 
+    blocks_in: Optional[RecordSpecification]
+    blocks_out: Optional[RecordSpecification]
+    strict: bool
+
     def __init__(
         self,
         *,
         blocks_in: Optional[RecordSpecification] = None,
-        blocks_out: Optional[RecordSpecification] = None
+        blocks_out: Optional[RecordSpecification] = None,
+        strict: bool = True,
     ):
         """
         Build the transform specification. At least one of blocks_in or blocks_out must not be None.
 
         :param blocks_in: incoming record specification, None for row-records.
         :param blocks_out: outgoing record specification, None for row-records.
+        :param strict: if True insist block be strict, and in and out blocks agree on row-form columns.∂
         """
+        assert isinstance(strict, bool)
+        self.strict = strict
         if blocks_in is not None:
             assert isinstance(blocks_in, RecordSpecification)
+            if strict:
+                assert blocks_in.strict
             ck = [k for k in blocks_in.content_keys if k is not None]
             if len(ck) != len(set(ck)):
                 raise ValueError("blocks_in can not have duplicate content keys")
         if blocks_out is not None:
             assert isinstance(blocks_out, RecordSpecification)
+            if strict:
+                assert blocks_out.strict
         if (blocks_in is None) and (blocks_out is None):
             raise ValueError(
                 "At least one of blocks_in or blocks_out should not be None"
@@ -388,6 +244,8 @@ class RecordMap:
             unknown = set(blocks_out.content_keys) - set(blocks_in.content_keys)
             if len(unknown) > 0:
                 raise ValueError("unknown outgoing content_keys" + str(unknown))
+            if strict:
+                assert set(blocks_out.record_keys) == set(blocks_in.record_keys)
         self.blocks_in = blocks_in
         self.blocks_out = blocks_out
         if self.blocks_in is not None:
@@ -432,30 +290,33 @@ class RecordMap:
         :param local_data_model: optional Pandas data model.
         :return: example result data frame.
         """
-        if local_data_model is None:
-            local_data_model = data_algebra.pandas_model.default_data_model
         if self.blocks_in is not None:
-            example = self.blocks_in.control_table.copy()
-            nrow = example.shape[0]
-            for rk in self.blocks_in.record_keys:
-                example[rk] = [rk] * nrow
+            if local_data_model is None:
+                local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(self.blocks_in.control_table)
+            assert isinstance(local_data_model, data_algebra.data_model.DataModel)
+            example_data = local_data_model.clean_copy(self.blocks_in.control_table)
+            if (self.blocks_in.record_keys is None) or (len(self.blocks_in.record_keys) == 0):
+                return example_data
+            nrow = example_data.shape[0]
+            row_key_frame = local_data_model.data_frame({k: [k] * nrow for k in self.blocks_in.record_keys})
+            example = local_data_model.concat_columns([row_key_frame, example_data])
             return example
-        if self.blocks_in is not None:
-            example = local_data_model.data_frame()
-            for k in self.blocks_out.row_columns:
-                example[k] = [k]
+        if self.blocks_out is not None:
+            if local_data_model is None:
+                local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(self.blocks_out.control_table)
+            assert isinstance(local_data_model, data_algebra.data_model.DataModel)
+            example = local_data_model.data_frame({k : [k] for k in self.blocks_out.row_columns})
             return example
         return None
 
     # noinspection PyPep8Naming
     def transform(
-        self, X, *, check_blocks_out_keying: bool = False, local_data_model=None
+        self, X, *, local_data_model=None
     ):
         """
         Transform X records.
 
         :param X: data frame to be transformed.
-        :param check_blocks_out_keying: logical, if True check output key constraints.
         :param local_data_model: pandas data model.
         :return: transformed data frame.
         """
@@ -463,18 +324,18 @@ class RecordMap:
         if len(unknown) > 0:
             raise ValueError("missing required columns: " + str(unknown))
         if local_data_model is None:
-            local_data_model = data_algebra.pandas_model.default_data_model
-        X = X.reset_index(drop=True)
+            local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(X)
+        assert isinstance(local_data_model, data_algebra.data_model.DataModel)
+        assert local_data_model.is_appropriate_data_instance(X)
+        X = local_data_model.clean_copy(X)
         if self.blocks_in is not None:
-            X = blocks_to_rowrecs(
-                X, blocks_in=self.blocks_in, local_data_model=local_data_model
+            X = local_data_model.blocks_to_rowrecs(
+                X, blocks_in=self.blocks_in
             )
         if self.blocks_out is not None:
-            X = rowrecs_to_blocks(
+            X = local_data_model.rowrecs_to_blocks(
                 X,
                 blocks_out=self.blocks_out,
-                check_blocks_out_keying=check_blocks_out_keying,
-                local_data_model=local_data_model,
             )
         return X
 
@@ -498,6 +359,7 @@ class RecordMap:
         out = s2.transform(s1.transform(inp))
         rsi = inp.drop(rk, axis=1, inplace=False)
         rso = out.drop(rk, axis=1, inplace=False)
+        strict = self.strict and other.strict
         if inp.shape[0] < 2:
             if out.shape[0] < 2:
                 return None
@@ -507,7 +369,9 @@ class RecordMap:
                         control_table=rso,
                         record_keys=rk,
                         control_table_keys=s2.blocks_out.control_table_keys,
-                    )
+                        strict=strict,
+                    ),
+                    strict=strict,
                 )
         else:
             if out.shape[0] < 2:
@@ -516,7 +380,9 @@ class RecordMap:
                         control_table=rsi,
                         record_keys=rk,
                         control_table_keys=s1.blocks_in.control_table_keys,
-                    )
+                        strict=strict,
+                    ),
+                    strict=strict,
                 )
             else:
                 return RecordMap(
@@ -524,28 +390,23 @@ class RecordMap:
                         control_table=rsi,
                         record_keys=rk,
                         control_table_keys=s1.blocks_in.control_table_keys,
+                        strict=strict,
                     ),
                     blocks_out=RecordSpecification(
                         control_table=rso,
                         record_keys=rk,
                         control_table_keys=s2.blocks_out.control_table_keys,
+                        strict=strict,
                     ),
+                    strict=strict,
                 )
 
-    # noinspection PyTypeChecker
-    def __rrshift__(self, other):  # override other >> self
-        if other is None:
-            return self
-        if isinstance(other, RecordMap):
-            # (data >> other) >> self == data >> (other >> self)
-            return self.compose(other)
-        return self.transform(other)
-
     def inverse(self):
         """
-        Return inverse transform.
+        Return inverse transform, if there is such (duplicate value keys or mis-matching
+        row representations can prevent this).
         """
-        return RecordMap(blocks_in=self.blocks_out, blocks_out=self.blocks_in)
+        return RecordMap(blocks_in=self.blocks_out, blocks_out=self.blocks_in, strict=self.strict)
 
     def fmt(self) -> str:
         """Format for informal presentation."""
@@ -594,6 +455,8 @@ class RecordMap:
             + self.blocks_in.__repr__()
             + ",\n    blocks_out="
             + self.blocks_out.__repr__()
+            + ",\n    strict="
+            + self.strict.__repr__()
             + ")"
         )
         return s
@@ -653,7 +516,8 @@ def pivot_blocks_to_rowrecs(
     """
 
     if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
+        local_data_model = data_algebra.data_model.default_data_model()
+    assert isinstance(local_data_model, data_algebra.data_model.DataModel)
     control_table = local_data_model.data_frame(
         {
             attribute_key_column: record_value_columns,
@@ -675,7 +539,7 @@ def pivot_rowrecs_to_blocks(
     attribute_value_column,
     record_keys,
     record_value_columns,
-    local_data_model=None
+    local_data_model=None,
 ):
     """
     Build a row records to block records map. This is very similar to a SQL unpivot.
@@ -689,7 +553,8 @@ def pivot_rowrecs_to_blocks(
     """
 
     if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
+        local_data_model = data_algebra.data_model.default_data_model()
+    assert isinstance(local_data_model, data_algebra.data_model.DataModel)
     control_table = local_data_model.data_frame(
         {
             attribute_key_column: record_value_columns,
@@ -710,7 +575,8 @@ def pivot_specification(
     row_keys: Iterable[str],
     col_name_key: str = "column_name",
     col_value_key: str = "column_value",
-    value_cols: Iterable[str]
+    value_cols: Iterable[str],
+    local_data_model = None,
 ) -> RecordMap:
     """
     Specify the cdata transformation that pivots records from a single column of values into collected rows.
@@ -720,6 +586,7 @@ def pivot_specification(
     :param col_name_key: column name to take the names of columns as a column
     :param col_value_key: column name to take the values in columns as a column
     :param value_cols: columns to place values in
+    :param local_data_model: data.frame data model
     :return: RecordSpecification
     """
     assert not isinstance(value_cols, str)
@@ -730,9 +597,12 @@ def pivot_specification(
     assert isinstance(col_value_key, str)
     known_cols = row_keys + [col_name_key, col_value_key] + value_cols
     assert len(known_cols) == len(set(known_cols))
+    if local_data_model is None:
+        local_data_model = data_algebra.data_model.default_data_model()
+    assert isinstance(local_data_model, data_algebra.data_model.DataModel)
     record_map = RecordMap(
         blocks_in=RecordSpecification(
-            control_table=data_algebra.pandas_model.default_data_model.pd.DataFrame(
+            control_table=local_data_model.pd.DataFrame(
                 {
                     col_name_key: value_cols,
                     col_value_key: value_cols,
@@ -750,7 +620,8 @@ def unpivot_specification(
     row_keys: Iterable[str],
     col_name_key: str = "column_name",
     col_value_key: str = "column_value",
-    value_cols: Iterable[str]
+    value_cols: Iterable[str],
+    local_data_model=None,
 ) -> RecordMap:
     """
     Specify the cdata transformation that un-pivots records into a single column of values plus keys.
@@ -760,6 +631,7 @@ def unpivot_specification(
     :param col_name_key: column name to land the names of columns as a column
     :param col_value_key: column name to land the values in columns as a column
     :param value_cols: columns to take values from
+    :param local_data_model: data.frame data model
     :return: RecordSpecification
     """
     assert not isinstance(value_cols, str)
@@ -770,9 +642,12 @@ def unpivot_specification(
     assert isinstance(col_value_key, str)
     known_cols = row_keys + [col_name_key, col_value_key] + value_cols
     assert len(known_cols) == len(set(known_cols))
+    if local_data_model is None:
+        local_data_model = data_algebra.data_model.default_data_model()
+    assert isinstance(local_data_model, data_algebra.data_model.DataModel)
     record_map = RecordMap(
         blocks_out=RecordSpecification(
-            control_table=data_algebra.pandas_model.default_data_model.pd.DataFrame(
+            control_table=local_data_model.pd.DataFrame(
                 {
                     col_name_key: value_cols,
                     col_value_key: value_cols,
