@@ -3,7 +3,7 @@ Base class for adapters for Pandas-like APIs
 """
 
 from abc import ABC
-from typing import Any, Callable, Dict, List, Optional
+from typing import Any, Callable, Dict, Iterable, List, Optional
 import datetime
 import types
 import numbers
@@ -16,9 +16,11 @@ import data_algebra.data_model
 import data_algebra.expr_rep
 import data_algebra.data_ops_types
 import data_algebra.connected_components
+import data_algebra.cdata
+import data_algebra.expression_walker
 
 
-# TODO: possibly import dask, Nvidia Rapids, modin, datatable versions
+# also possible, Dask, Nvidia Rapids, Modin, or Datatable versions
 
 
 def none_mark_scalar_or_length(v) -> Optional[int]:
@@ -54,12 +56,8 @@ def promote_scalar_to_array(vi, *, target_len: int) -> List:
     # noinspection PyBroadException
     if len_v is None:
         return [vi] * target_len  # scalar
-    if len_v == target_len:
-        return vi
-    if len_v == 1:
-        return [vi[0]] * target_len  # TODO: see if we can eliminate this one
-    else:
-        raise ValueError("incompatible column lengths")
+    assert len_v == target_len
+    return vi
 
 
 def _negate_or_subtract(*args):
@@ -90,6 +88,10 @@ def _type_safe_not_equal(a, b):
     return numpy.not_equal(a, b)
 
 
+def _not_equal_false(a):
+    return a != False
+
+
 def _type_safe_is_in(a, b):
     b = list(b)
     if len(b) > 0:
@@ -143,7 +145,7 @@ def _where_expr(*args):
 
 
 # base class for Pandas-like API realization
-class PandasModelBase(data_algebra.data_model.DataModel, ABC):
+class PandasModelBase(data_algebra.data_model.DataModel, data_algebra.expression_walker.ExpressionWalker, ABC):
     """
     Base class for implementing the data algebra on pandas-like APIs
     """
@@ -157,26 +159,29 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
     def __init__(self, *, pd: types.ModuleType, presentation_model_name: str):
         assert isinstance(pd, types.ModuleType)
         data_algebra.data_model.DataModel.__init__(
-            self, presentation_model_name=presentation_model_name
+            self, presentation_model_name=presentation_model_name, module=pd
+        )
+        data_algebra.expression_walker.ExpressionWalker.__init__(
+            self,
         )
         self.pd = pd
         self.impl_map = self._populate_impl_map()
         self.transform_op_map = {"any_value": "first"}
         self.user_fun_map = dict()
         self._method_dispatch_table = {
-            "ConcatRowsNode": self.concat_rows_step,
-            "ConvertRecordsNode": self.convert_records_step,
-            "DropColumnsNode": self.drop_columns_step,
-            "ExtendNode": self.extend_step,
-            "NaturalJoinNode": self.natural_join_step,
-            "OrderRowsNode": self.order_rows_step,
-            "ProjectNode": self.project_step,
-            "MapColumnsNode": self.map_columns_step,
-            "RenameColumnsNode": self.rename_columns_step,
-            "SelectColumnsNode": self.select_columns_step,
-            "SelectRowsNode": self.select_rows_step,
-            "SQLNode": self.sql_proxy_step,
-            "TableDescription": self.table_step,
+            "ConcatRowsNode": self._concat_rows_step,
+            "ConvertRecordsNode": self._convert_records_step,
+            "DropColumnsNode": self._drop_columns_step,
+            "ExtendNode": self._extend_step,
+            "NaturalJoinNode": self._natural_join_step,
+            "OrderRowsNode": self._order_rows_step,
+            "ProjectNode": self._project_step,
+            "MapColumnsNode": self._map_columns_step,
+            "RenameColumnsNode": self._rename_columns_step,
+            "SelectColumnsNode": self._select_columns_step,
+            "SelectRowsNode": self._select_rows_step,
+            "SQLNode": self._sql_proxy_step,
+            "TableDescription": self._table_step,
         }
 
     # implementations
@@ -288,12 +293,12 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             "%": numpy.mod,
             "**": numpy.power,
             "and": _k_and,
-            "&": numpy.bitwise_and,
+            "&": _k_and,
             "or": _k_or,
             "|": _k_or,
             "xor": numpy.logical_xor,
             "^": numpy.logical_xor,
-            "not": numpy.bitwise_or,
+            "not": _not_equal_false,
             "where": _where_expr,
             "if_else": lambda *args: self._if_else_expr(*args),
             "is_nan": self.isnan,
@@ -384,6 +389,26 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         """
         # noinspection PyUnresolvedReferences
         return isinstance(df, self.pd.DataFrame)
+    
+    def clean_copy(self, df):
+        """
+        Copy of data frame without indices.
+        """
+        assert self.is_appropriate_data_instance(df)
+        return df.reset_index(drop=True, inplace=False)
+    
+    def to_pandas(self, df):
+        """
+        Convert to Pandas
+        """
+        return self.clean_copy(df)
+    
+    def drop_indices(self, df) -> None:
+        """
+        Drop indices in place.
+        """
+        assert self.is_appropriate_data_instance(df)
+        df.reset_index(drop=True, inplace=True)
 
     def can_convert_col_to_numeric(self, x):
         """
@@ -431,11 +456,57 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                 self.pd.isnull(x), numpy.logical_or(numpy.isnan(x), numpy.isinf(x))
             )
         return self.pd.isnull(x)
+    
+    def concat_rows(self, frame_list: List):
+        """
+        Concatenate rows from frame_list
+        """
+        frame_list = list(frame_list)
+        assert len(frame_list) > 0
+        if len(frame_list) == 1:
+            return self.clean_copy(frame_list[0])
+        res = self.pd.concat(frame_list, axis=0)
+        return res
+
+    def concat_columns(self, frame_list: List):
+        """
+        Concatenate columns from frame_list
+        """
+        frame_list = list(frame_list)
+        assert len(frame_list) > 0
+        if len(frame_list) == 1:
+            return self.clean_copy(frame_list[0])
+        res = self.pd.concat(frame_list, axis=1)
+        return res
+
+    def table_is_keyed_by_columns(self, table, *, column_names: Iterable[str]) -> bool:
+        """
+        Check if a table is keyed by a given list of column names.
+
+        :param table: DataFrame
+        :param column_names: list of column names
+        :return: True if rows are uniquely keyed by values in named columns
+        """
+        # check for ill-condition
+        if isinstance(column_names, str):
+            column_names = [column_names]
+        else:
+            column_names = list(column_names)
+        missing_columns = set(column_names) - set(table.columns)
+        if len(missing_columns) > 0:
+            return False
+        # get rid of some corner cases
+        if table.shape[0] < 2:
+            return True
+        if len(column_names) < 1:
+            return False
+        counts = table.groupby(column_names, observed=True).size()
+        return max(counts) <= 1
 
     # bigger stuff
 
     # noinspection PyMethodMayBeStatic,PyUnusedLocal
-    def table_step(self, op, *, data_map: dict, narrow: bool):
+    def _table_step(self, op, *, data_map: dict):
         """
         Return a copy of data frame from table description and data_map.
         """
@@ -458,23 +529,24 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                 )
         # check all columns we expect are present
         columns_using = op.column_names
-        if not narrow:
-            columns_using = [c for c in df.columns]
-        missing = set(columns_using) - set([c for c in df.columns])
+        missing = set(columns_using) - set(df.columns)
         if len(missing) > 0:
             raise ValueError("missing required columns: " + str(missing))
         # make an index-free copy of the data to isolate side-effects and not deal with indices
         res = df.loc[:, columns_using]
-        res = res.reset_index(drop=True, inplace=False)  # copy and clear out indices
+        res = self.clean_copy(res)
         return res
 
-    def sql_proxy_step(self, op, *, data_map: dict, narrow: bool):
+    def _sql_proxy_step(self, op, *, data_map: dict):
         """
         execute SQL
         """
         assert op.node_name == "SQLNode"
         db_handle = data_map[op.view_name]
+        # would like (but causes cicrular import) assert isinstance(db_handle, data_algebra.db_model.DBHandle)
         res = db_handle.read_query("\n".join(op.sql))
+        res = self.data_frame(res[op.columns_produced()])
+        assert self.is_appropriate_data_instance(res)
         return res
 
     def columns_to_frame_(self, cols: Dict[str, Any], *, target_rows: Optional[int] = None):
@@ -494,7 +566,9 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if len(cols) < 1:
             # all scalars, so nothing carrying index information
             if target_rows is not None:
-                return self.pd.DataFrame({}, index=range(target_rows)).reset_index(drop=True, inplace=False)
+                res = self.pd.DataFrame({}, index=range(target_rows))
+                self.drop_indices(res)
+                return res
             else:
                 return self.pd.DataFrame({})
         was_all_scalars = True
@@ -511,7 +585,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                 target_rows = 1
             # all scalars, so nothing carrying index information
             promoted_cols = {k: promote_scalar_to_array(v, target_len=target_rows) for (k, v) in cols.items()}
-            return self.pd.DataFrame(promoted_cols, index=range(target_rows)).reset_index(drop=True, inplace=False)
+            return self.clean_copy(self.pd.DataFrame(promoted_cols, index=range(target_rows)))
         assert target_rows is not None
         if target_rows < 1:
             # no rows, so presuming no index information (shouldn't have come from an aggregation)
@@ -530,9 +604,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if (res.shape[0] == 0) and (transient_new_frame.shape[0] > 0):
             # scalars get interpreted as single row items, instead of zero row items
             # growing the extension frame
-            transient_new_frame = transient_new_frame.iloc[range(0), :].reset_index(
-                drop=True
-            )
+            transient_new_frame = self.clean_copy(transient_new_frame.iloc[range(0), :])
         if res.shape[0] == transient_new_frame.shape[0]:
             if res.shape[1] < 1:
                 return transient_new_frame
@@ -549,33 +621,33 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             res[c] = transient_new_frame[c]
         return res
 
-    def eval(self, op, *, data_map: Dict[str, Any], narrow: bool = False):
+    def eval(self, op, *, data_map: Dict[str, Any]):
         """
         Implementation of Pandas evaluation of operators
 
         :param op: ViewRepresentation to evaluate
         :param data_map: dictionary mapping table and view names to data frames or data sources
-        :param narrow: if True narrow results to only columns anticipated
         :return: data frame result
         """
         assert isinstance(data_map, Dict)
-        return self._eval_value_source(s=op, data_map=data_map, narrow=narrow)
+        assert isinstance(op, data_algebra.data_ops_types.OperatorPlatform)
+        return self._eval_value_source(s=op, data_map=data_map)
 
-    def _eval_value_source(self, s, *, data_map: dict, narrow: bool):
+    def _eval_value_source(self, s, *, data_map: dict):
         """
         Evaluate an incoming (or value source) node.
         """
         return self._method_dispatch_table[s.node_name](
-            op=s, data_map=data_map, narrow=narrow
+            op=s, data_map=data_map
         )
 
-    def extend_step(self, op, *, data_map, narrow):
+    def _extend_step(self, op, *, data_map):
         """
         Execute an extend step, returning a data frame.
         """
         if op.node_name != "ExtendNode":
             raise TypeError("op was supposed to be a data_algebra.data_ops.ExtendNode")
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         if res.shape[0] <= 0:
             # special case out no-row frame
             incoming_col_set = set(res.columns)
@@ -596,7 +668,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                 warnings.simplefilter(
                     "ignore"
                 )  # out of range things like arccosh were warning
-                new_cols = {k: opk.evaluate(res) for k, opk in op.ops.items()}
+                new_cols = {k: opk.act_on(res, expr_walker=self) for k, opk in op.ops.items()}
             new_frame = self.columns_to_frame_(new_cols, target_rows=res.shape[0])
             res = self.add_data_frame_columns_to_data_frame_(res, new_frame)
         else:
@@ -607,7 +679,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             col_set = set(col_list)
             for c in op.order_by:
                 if c not in col_set:
-                    col_list = col_list + [c]
+                    col_list.append(c)
                     col_set.add(c)
             order_cols = [c for c in col_list]  # must be partition by followed by order
             for (k, opk) in op.ops.items():
@@ -616,7 +688,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     if isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference):
                         value_name = opk.args[0].column_name
                         if value_name not in col_set:
-                            col_list = col_list + [value_name]
+                            col_list.append(value_name)
                             col_set.add(value_name)
                     elif isinstance(opk.args[0], data_algebra.expr_rep.Value):
                         key = str(opk.args[0].value)
@@ -630,19 +702,19 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     else:
                         raise ValueError("opk must be a ColumnReference or Value")
             ascending = [c not in set(op.reverse) for c in col_list]
-            subframe = res[col_list].reset_index(drop=True)
+            subframe = self.clean_copy(res[col_list])
             subframe["_data_algebra_orig_index"] = subframe.index
             if len(order_cols) > 0:
-                subframe = subframe.sort_values(
+                subframe = self.clean_copy(subframe.sort_values(
                     by=col_list, ascending=ascending
-                ).reset_index(drop=True)
+                ))
             subframe[standin_name] = 1
             if len(op.partition_by) > 0:
-                opframe = subframe.groupby(op.partition_by)
+                opframe = subframe.groupby(op.partition_by, observed=True)
                 #  Groupby preserves the order of rows within each group.
                 # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.groupby.html
             else:
-                opframe = subframe.groupby([standin_name])
+                opframe = subframe.groupby([standin_name], observed=True)
             # perform calculations
             for (k, opk) in op.ops.items():
                 # work on a slice of the data frame
@@ -686,7 +758,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     if isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference):
                         value_name = opk.args[0].column_name
                         if value_name not in set(col_list):
-                            col_list = col_list + [value_name]
+                            col_list.append(value_name)
                         transform_op = opk.op
                         try:
                             transform_op = self.transform_op_map[transform_op]
@@ -715,11 +787,11 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             # copy out results
             subframe = subframe.sort_values(by=["_data_algebra_orig_index"])
             subframe = subframe.loc[:, list(op.ops.keys())]
-            subframe = subframe.reset_index(drop=True)
+            subframe = self.clean_copy(subframe)
             res = self.add_data_frame_columns_to_data_frame_(res, subframe)
         return res
 
-    def project_step(self, op, *, data_map, narrow):
+    def _project_step(self, op, *, data_map):
         """
         Execute a project step, returning a data frame.
         """
@@ -731,7 +803,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         # try the following tutorial:
         # https://www.shanelynn.ie/summarising-aggregation-and-grouping-data-in-python-pandas/
         data_algebra_temp_cols = {}
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         for (k, opk) in op.ops.items():
             if len(opk.args) > 1:
                 raise ValueError(
@@ -757,7 +829,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     )
         res["_data_table_temp_col"] = 1
         if len(op.group_by) > 0:
-            res = res.groupby(op.group_by)
+            res = res.groupby(op.group_by, observed=True)
         if len(op.ops) > 0:
             cols = {}
             for k, opk in op.ops.items():
@@ -791,7 +863,8 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         # agg can return scalars, which then can't be made into a self.pd.DataFrame
         res = self.columns_to_frame_(cols)
         res = res.reset_index(
-            drop=(len(op.group_by) < 1) or (res.shape[0] <= 0)
+            drop=(len(op.group_by) < 1) or (res.shape[0] <= 0),
+            inplace=False,
         )  # grouping variables in the index
         missing_group_cols = set(op.group_by) - set(res.columns)
         if res.shape[0] > 0:
@@ -803,11 +876,11 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if "_data_table_temp_col" in res.columns:
             res = res.drop("_data_table_temp_col", axis=1, inplace=False)
         # double check shape is what we expect
-        if not data_algebra.util.table_is_keyed_by_columns(res, op.group_by):
+        if not self.table_is_keyed_by_columns(res, column_names=op.group_by):
             raise ValueError("result wasn't keyed by group_by columns")
         return res
 
-    def select_rows_step(self, op, *, data_map, narrow):
+    def _select_rows_step(self, op, *, data_map):
         """
         Execute a select rows step, returning a data frame.
         """
@@ -815,14 +888,14 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.SelectRowsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         if res.shape[0] < 1:
             return res
-        selection = op.expr.evaluate(res)
-        res = res.loc[selection, :].reset_index(drop=True, inplace=False)
+        selection = op.expr.act_on(res, expr_walker=self)
+        res = self.clean_copy(res.loc[selection, :])
         return res
 
-    def select_columns_step(self, op, *, data_map, narrow):
+    def _select_columns_step(self, op, *, data_map):
         """
         Execute a select columns step, returning a data frame.
         """
@@ -830,10 +903,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.SelectColumnsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         return res[op.column_selection]
 
-    def drop_columns_step(self, op, *, data_map, narrow):
+    def _drop_columns_step(self, op, *, data_map):
         """
         Execute a drop columns step, returning a data frame.
         """
@@ -841,11 +914,11 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.DropColumnsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         column_selection = [c for c in res.columns if c not in op.column_deletions]
         return res[column_selection]
 
-    def order_rows_step(self, op, *, data_map, narrow):
+    def _order_rows_step(self, op, *, data_map):
         """
         Execute an order rows step, returning a data frame.
         """
@@ -853,19 +926,18 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.OrderRowsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         if res.shape[0] > 1:
             ascending = [
                 False if ci in set(op.reverse) else True for ci in op.order_columns
             ]
-            res = res.sort_values(by=op.order_columns, ascending=ascending).reset_index(
-                drop=True
-            )
+            res = res.sort_values(by=op.order_columns, ascending=ascending, ignore_index=True, inplace=False)
+            self.drop_indices(res)
         if (op.limit is not None) and (res.shape[0] > op.limit):
-            res = res.iloc[range(op.limit), :].reset_index(drop=True)
+            res = self.clean_copy(res.iloc[range(op.limit), :])
         return res
 
-    def map_columns_step(self, op, *, data_map, narrow):
+    def _map_columns_step(self, op, *, data_map):
         """
         Execute a map columns step, returning a data frame.
         """
@@ -873,14 +945,14 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.MapColumnsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         res = res.rename(columns=op.column_remapping)
         if (op.column_deletions is not None) and (len(op.column_deletions) > 0):
             column_selection = [c for c in res.columns if c not in op.column_deletions]
             res = res[column_selection]
         return res
 
-    def rename_columns_step(self, op, *, data_map, narrow):
+    def _rename_columns_step(self, op, *, data_map):
         """
         Execute a rename columns step, returning a data frame.
         """
@@ -888,7 +960,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.RenameColumnsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         return res.rename(columns=op.reverse_mapping)
 
     # noinspection PyMethodMayBeStatic
@@ -908,7 +980,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             pass
         return jointype
 
-    def natural_join_step(self, op, *, data_map, narrow):
+    def _natural_join_step(self, op, *, data_map):
         """
         Execute a natural join step, returning a data frame.
         """
@@ -916,8 +988,8 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.NaturalJoinNode"
             )
-        left = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
-        right = self._eval_value_source(op.sources[1], data_map=data_map, narrow=narrow)
+        left = self._eval_value_source(op.sources[0], data_map=data_map)
+        right = self._eval_value_source(op.sources[1], data_map=data_map)
         if (left.shape[0] == 0) and (right.shape[0] == 0):
             # pandas seems to not like this case
             return self.pd.DataFrame({k: [] for k in op.columns_produced()})
@@ -948,7 +1020,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             sort=False,
             suffixes=("", "_tmp_right_col"),
         )
-        res = res.reset_index(drop=True)
+        self.drop_indices(res)
         if scratch_col is not None:
             del res[scratch_col]
         on_a_set = set(op.on_a)
@@ -957,10 +1029,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                 is_null = res[c].isnull()
                 res.loc[is_null, c] = res.loc[is_null, c + "_tmp_right_col"]
                 res = res.drop(c + "_tmp_right_col", axis=1, inplace=False)
-        res = res.reset_index(drop=True)
+        self.drop_indices(res)
         return res
 
-    def concat_rows_step(self, op, *, data_map, narrow):
+    def _concat_rows_step(self, op, *, data_map):
         """
         Execute a concat rows step, returning a data frame.
         """
@@ -968,8 +1040,8 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.ConcatRowsNode"
             )
-        left = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
-        right = self._eval_value_source(op.sources[1], data_map=data_map, narrow=narrow)
+        left = self._eval_value_source(op.sources[0], data_map=data_map)
+        right = self._eval_value_source(op.sources[1], data_map=data_map)
         if op.id_column is not None:
             if left.shape[0] > 0:
                 left[op.id_column] = op.a_name
@@ -988,10 +1060,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise ValueError(f"concat: incompatible column types: {type_checks}")
         # noinspection PyUnresolvedReferences
         res = self.pd.concat([left, right], axis=0, ignore_index=True, sort=False)
-        res = res.reset_index(drop=True)
+        self.drop_indices(res)
         return res
 
-    def convert_records_step(self, op, *, data_map, narrow):
+    def _convert_records_step(self, op, *, data_map):
         """
         Execute record conversion step, returning a data frame.
         """
@@ -999,5 +1071,238 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise TypeError(
                 "op was supposed to be a data_algebra.data_ops.ConvertRecordsNode"
             )
-        res = self._eval_value_source(op.sources[0], data_map=data_map, narrow=narrow)
+        res = self._eval_value_source(op.sources[0], data_map=data_map)
         return op.record_map.transform(res, local_data_model=self)
+    
+    # cdata record conversion steps
+
+    def blocks_to_rowrecs(self, data, *, blocks_in: data_algebra.cdata.RecordSpecification):
+        """
+        Convert a block record (record spanning multiple rows) into a rowrecord (record in a single row).
+
+        :param data: data frame to be transformed
+        :param blocks_in: record specification
+        :return: transformed data frame
+        """
+        assert isinstance(blocks_in, data_algebra.cdata.RecordSpecification)
+        ck = [k for k in blocks_in.content_keys if k is not None]
+        if len(ck) != len(set(ck)):
+            raise ValueError("blocks_in can not have duplicate content keys")
+        data = self.clean_copy(data[blocks_in.block_columns])
+        assert set(data.columns) == set(blocks_in.block_columns)
+        # table must be keyed by record_keys + control_table_keys
+        if not self.table_is_keyed_by_columns(
+            data, column_names=blocks_in.record_keys + blocks_in.control_table_keys
+        ):
+            raise ValueError(
+                "table is not keyed by blocks_in.record_keys + blocks_in.control_table_keys"
+            )
+        # convert to row-records
+        # regularize/complete records
+        dtemp = data.copy()  # TODO: select down columns
+        dtemp["FALSE_AGG_KEY"] = 1
+        if len(blocks_in.record_keys) > 0:
+            ideal = dtemp[blocks_in.record_keys + ["FALSE_AGG_KEY"]].copy()
+            res = ideal.groupby(blocks_in.record_keys, observed=True)["FALSE_AGG_KEY"].agg("sum")
+            ideal = self.data_frame(res).reset_index(drop=False, inplace=False)
+            ideal["FALSE_AGG_KEY"] = 1
+            ctemp = blocks_in.control_table[blocks_in.control_table_keys].copy()
+            ctemp["FALSE_AGG_KEY"] = 1
+            ideal = ideal.merge(ctemp, how="outer", on="FALSE_AGG_KEY")
+            self.drop_indices(ideal)
+            dtemp = ideal.merge(
+                right=dtemp,
+                how="left",
+                on=blocks_in.record_keys + blocks_in.control_table_keys + ["FALSE_AGG_KEY"],
+            )
+        dtemp = dtemp.sort_values(
+            by=blocks_in.record_keys + blocks_in.control_table_keys, inplace=False, ignore_index=True
+        )
+        self.drop_indices(dtemp)
+        # start building up result frame
+        if len(blocks_in.record_keys) > 0:
+            res = dtemp.groupby(blocks_in.record_keys, observed=True)["FALSE_AGG_KEY"].agg("sum")
+        else:
+            res = dtemp.groupby("FALSE_AGG_KEY", observed=True)["FALSE_AGG_KEY"].agg("sum")
+        res = self.data_frame(res).reset_index(drop=False, inplace=False)
+        res = res.sort_values(by=blocks_in.record_keys, inplace=False, ignore_index=True)
+        self.drop_indices(res)
+        del res["FALSE_AGG_KEY"]
+        # now fill in columns
+        ckeys = blocks_in.control_table_keys
+        value_keys = [k for k in blocks_in.control_table.columns if k not in set(ckeys)]
+        donor_cols = set(dtemp.columns)
+        for i in range(blocks_in.control_table.shape[0]):
+            want = numpy.ones((dtemp.shape[0],), dtype=bool)
+            for ck in ckeys:
+                want = numpy.logical_and(want, dtemp[ck] == blocks_in.control_table[ck][i])
+            if numpy.any(want):
+                for vk in value_keys:
+                    if vk in donor_cols:
+                        dcol = blocks_in.control_table[vk][i]
+                        res[dcol] = numpy.asarray(dtemp.loc[want, vk])
+        # fill in any missed columns
+        colset = set(res.columns)
+        for c in blocks_in.row_version():
+            if c not in colset:
+                res[c] = None
+        if data.shape[0] <= 0:
+            res = res.loc[range(0), :]
+            res = self.clean_copy(res)
+        return res
+
+    def rowrecs_to_blocks(
+        self,
+        data,
+        *,
+        blocks_out: data_algebra.cdata.RecordSpecification,
+    ):
+        """
+        Convert rowrecs (single row records) into block records (multiple row records).
+
+        :param data: data frame to transform.
+        :param blocks_out: record specification.
+        :return: transformed data frame
+        """
+        assert isinstance(blocks_out, data_algebra.cdata.RecordSpecification)
+        data = self.clean_copy(data[blocks_out.row_columns])
+        assert set(data.columns) == set(blocks_out.row_columns)
+        missing_cols = set(blocks_out.record_keys) - set(data.columns)
+        if len(missing_cols) > 0:
+            raise KeyError("missing required columns: " + str(missing_cols))
+        # table must be keyed by record_keys
+        if (data.shape[0] > 1) and (not self.table_is_keyed_by_columns(
+            data, column_names=blocks_out.record_keys
+        )):
+            raise ValueError(
+                "table is not keyed by blocks_out.record_keys"
+            )
+        # convert to block records, first build up parallel structures
+        rv = [k for k in blocks_out.row_version(include_record_keys=True) if k is not None]
+        if len(rv) != len(set(rv)):
+            raise ValueError("duplicate row columns")
+        dtemp_cols = [
+            k
+            for k in rv
+            if k is not None and k in set(blocks_out.record_keys + blocks_out.content_keys)
+        ]
+        dtemp = data[dtemp_cols].copy()
+        dtemp.sort_values(by=blocks_out.record_keys, inplace=True, ignore_index=True)
+        self.drop_indices(dtemp)
+        if len(dtemp.columns) != len(set(dtemp.columns)):
+            raise ValueError("targeted data columns not unique")
+        ctemp = blocks_out.control_table.copy()
+        dtemp["FALSE_JOIN_KEY"] = 1
+        ctemp["FALSE_JOIN_KEY"] = 1
+        res = dtemp[blocks_out.record_keys + ["FALSE_JOIN_KEY"]].merge(
+            ctemp, how="outer", on=["FALSE_JOIN_KEY"]
+        )
+        del res["FALSE_JOIN_KEY"]
+        ckeys = blocks_out.control_table_keys
+        res = res.sort_values(by=blocks_out.record_keys + ckeys, inplace=False, ignore_index=True)
+        self.drop_indices(res)
+        del ctemp["FALSE_JOIN_KEY"]
+        del dtemp["FALSE_JOIN_KEY"]
+        value_keys = [k for k in ctemp.columns if k not in set(ckeys)]
+        donor_cols = set(dtemp.columns)
+        for vk in value_keys:
+            res[vk] = None
+        # we now have parallel structures to copy between
+        for i in range(ctemp.shape[0]):
+            want = numpy.ones((res.shape[0],), dtype=bool)
+            for ck in ckeys:
+                want = numpy.logical_and(want, res[ck] == ctemp[ck][i])
+            if numpy.any(want):
+                for vk in value_keys:
+                    dcol = ctemp[vk][i]
+                    if dcol in donor_cols:
+                        nvals = numpy.asarray(dtemp[dcol])
+                        if len(nvals) < 1:
+                            nvals = [None] * numpy.sum(want)
+                        if numpy.all(want):
+                            res[vk] = nvals  # get around Pandas future warning
+                        else:
+                            res.loc[want, vk] = nvals
+        # see about promoting composite columns to numeric
+        for vk in set(value_keys):
+            converted = self.to_numeric(res[vk], errors="coerce")
+            if numpy.all(
+                self.isnull(converted) == self.isnull(res[vk])
+            ):
+                res[vk] = converted
+        if data.shape[0] < 1:
+            # empty input produces emtpy output (with different column structure)
+            res = self.clean_copy(res.iloc[range(0), :])
+        if data.shape[0] <= 0:
+            res = res.loc[range(0), :]
+            res = self.clean_copy(res)
+        return res
+
+    # expression helpers
+    
+    def act_on_literal(self, *, value):
+        """
+        Action for a literal/constant in an expression.
+
+        :param value: literal value being supplied
+        :return: converted result
+        """
+        return value
+    
+    def act_on_column_name(self, *, arg, value):
+        """
+        Action for a column name.
+
+        :param arg: item we are acting on
+        :param value: column name
+        :return: arg acted on
+        """
+        return arg[value]
+    
+    def act_on_expression(self, *, arg, values: List, op):
+        """
+        Action for a column name.
+
+        :param arg: item we are acting on
+        :param values: list of values to work on
+        :param op: operator to apply
+        :return: arg acted on
+        """
+        assert isinstance(values, List)
+        assert isinstance(op, data_algebra.expr_rep.Expression)
+        op_name = op.op
+        # check user fns
+        # first check chosen mappings
+        try:
+            method_to_call = self.user_fun_map[op_name]
+            return method_to_call(*values)
+        except KeyError:
+            pass
+        # check chosen mappings
+        try:
+            method_to_call = self.impl_map[op_name]
+            return method_to_call(*values)
+        except KeyError:
+            pass
+        # special zero argument functions
+        if len(values) == 0:
+            if op_name in ["uniform", "_uniform"]:
+                return numpy.random.uniform(size=arg.shape[0])
+            else:
+                KeyError(f"zero-argument function {op_name} not found")
+        # now see if argument (usually Pandas) can do this
+        # doubt we hit in this, as most exposed methods are window methods
+        try:
+            method = getattr(values[0], op_name)
+            if callable(method):
+                return method(*values[1:])
+        except AttributeError:
+            pass
+        # now see if numpy can do this
+        try:
+            fn = numpy.__dict__[op_name]
+            if callable(fn):
+                return fn(*values)
+        except KeyError:
+            pass
+        raise KeyError(f"function {op_name} not found")
